


__api_version__ = 2
__author__ = "Jürgen Knauth"
__version__ = "0.2022.10.25"



from .FileUncompressionGuess import FileUncompressionGuess

from .SpoolInfo import SpoolInfo
from .Spooler import Spooler

from .Packer import Packer
from .Packer1 import Packer1
from .Unpacker import Unpacker
from .Unpacker1 import Unpacker1
